from .datparser import DATParser
from .dupeparser import DupeParser
from .gamefinder import GameFinder
from .romchooser import ROMChooser
from .romdownloader import ROMDownloader
from .rommover import ROMMover
from .romparser import ROMParser
from .romsearch import ROMSearch

__all__ = [
    "DATParser",
    "DupeParser",
    "GameFinder",
    "ROMChooser",
    "ROMDownloader",
    "ROMMover",
    "ROMParser",
    "ROMSearch",
]



from .discord import discord_push
from .general import split, get_parent_name, get_file_time
from .io import load_yml, save_yml, unzip_file, load_json, save_json
from .logger import setup_logger
from .regex_matching import get_file_pattern, get_bracketed_file_pattern, get_game_name, get_short_name

__all__ = [
    "setup_logger",
    "load_yml",
    "save_yml",
    "get_bracketed_file_pattern",
    "get_file_pattern",
    "get_game_name",
    "get_short_name",
    "load_json",
    "save_json",
    "unzip_file",
    "discord_push",
    "split",
    "get_parent_name",
    "get_file_time",
]



from PySide6.QtWidgets import QDialog

from .layout_about import Ui_About


class AboutWindow(QDialog):

    def __init__(self, parent=None):
        """ROMSearch About window"""

        super().__init__()

        self.ui = Ui_About()
        self.ui.setupUi(self)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'romsearch'
copyright = '2024, bbtufty'
author = 'bbtufty'
release = '0.0.5.post2'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.napoleon',
              'sphinx.ext.todo',
              'sphinx.ext.viewcode',
              'sphinx_automodapi.automodapi',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

master_doc = 'index'

todo_include_todos = True

html_theme_options = {
    'collapse_navigation': False,
    'navigation_depth': 4,
    'globaltoc_collapse': False,
    'globaltoc_includehidden': False,
    'display_version': True,
}

autoclass_content = 'both'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = []



import json
import os
import zipfile

import yaml


class DumperEdit(yaml.Dumper):

    def increase_indent(self, flow=False, indentless=False):
        return super(DumperEdit, self).increase_indent(flow, False)

    def write_line_break(self, data=None):
        super().write_line_break(data)

        if len(self.indents) == 1:
            super().write_line_break()


def load_yml(f):
    """Load YAML file"""

    with open(f, "r") as file:
        config = yaml.safe_load(file)

    return config


def save_yml(f, data):
    """Save YAML file"""

    with open(f, "w") as file:
        yaml.dump(data,
                  file,
                  Dumper=DumperEdit,
                  default_flow_style=False,
                  sort_keys=False,
                  indent=2,
                  )


def unzip_file(zip_file_name,
               out_dir,
               ):
    """Unzip a file"""

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    out_dir = str(out_dir)

    with zipfile.ZipFile(zip_file_name, 'r') as zip_file:
        zip_file.extractall(out_dir)

    return True


def load_json(file):
    """Load json file"""

    with open(file, "r", encoding="utf-8") as f:
        j = json.load(f)

    return j


def save_json(data, out_file):
    """Save json in a pretty way"""

    with open(out_file, "w", encoding="utf-8") as f:
        json.dump(data,
                  f,
                  ensure_ascii=False,
                  indent=4,
                  )



import os
from functools import partial

from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (QMainWindow,
                               QFileDialog,
                               QWidget,
                               QGridLayout,
                               QHBoxLayout,
                               QVBoxLayout,
                               QLabel,
                               QFrame,
                               QPlainTextEdit,
                               )

import romsearch
from romsearch.util import load_yml, save_yml, load_json, save_json
from .gui_utils import add_item_to_list, get_gui_logger


def set_ordered_list(list_widget,
                     item_dict,
                     items,
                     ):
    """Set checked items from an ordered list

    Args:
        list_widget (QListWidget): The QListWidget
        item_dict (dict): A dictionary of items that contains the list items to index
        items (list): A list of items to set checked, in order
    """

    # Set the items to checked
    for i in items:
        item_dict[i]["item"].setCheckState(Qt.CheckState.Checked)
        item_dict[i]["check_state"] = True

    # Move the items into order
    for i in items[::-1]:
        idx = list_widget.row(item_dict[i]["item"])
        take_item = list_widget.takeItem(idx)
        list_widget.insertItem(0, take_item)
        list_widget.setCurrentRow(0)

    return True


def add_include_exclude_layout(tab_config,
                               text="Includes",
                               ):
    """Add the include/exclude table

    Args:
        tab_config (QWidget): The table widget to add the left/right tables to
        text (str): The text to put on top of the text edit
    """

    font = QFont()
    font.setBold(True)

    # Excludes
    hlayout = QHBoxLayout()
    vlayout = QVBoxLayout()

    label_exclude = QLabel(tab_config)
    label_exclude.setText(text)
    label_exclude.setFont(font)
    label_exclude.setAlignment(Qt.AlignmentFlag.AlignCenter)
    vlayout.addWidget(label_exclude)

    line = QFrame(tab_config)
    line.setFrameShadow(QFrame.Shadow.Plain)
    line.setFrameShape(QFrame.Shape.HLine)
    vlayout.addWidget(line)

    text_edit = QPlainTextEdit(tab_config)
    text_edit.setFrameShape(QFrame.Shape.WinPanel)
    text_edit.setFrameShadow(QFrame.Shadow.Plain)
    vlayout.addWidget(text_edit)

    hlayout.addLayout(vlayout)

    return text_edit, hlayout


class ConfigWindow(QMainWindow):

    def __init__(self,
                 main_ui,
                 logger=None,
                 ):
        """Class for the configuration window"""

        super().__init__()

        if logger is None:
            logger = get_gui_logger()
        self.logger = logger

        self.ui = main_ui

        # Read in the various pre-set configs we've got
        self.mod_dir = os.path.dirname(romsearch.__file__)

        default_file = os.path.join(self.mod_dir, "configs", "defaults.yml")
        default_config = load_yml(default_file)
        self.default_config = default_config

        self.config = {}

        self.include_exclude_dict = {}

        # Set up platforms
        self.platform_items = self.populate_list("platforms",
                                                 self.ui.listWidgetConfigPlatforms,
                                                 check_state=False,
                                                 )

        self.ui.listWidgetConfigPlatforms.itemChanged.connect(self.get_platforms_changed)

        # Set up regions/languages
        self.region_items = self.populate_list("regions",
                                               self.ui.listWidgetConfigRegionsLanguagesRegions,
                                               check_state=False,
                                               )
        self.language_items = self.populate_list("languages",
                                                 self.ui.listWidgetConfigRegionsLanguagesLanguages,
                                                 check_state=False,
                                                 )

        # Set up groupings for later
        self.all_dirs = {
            "raw_dir": self.ui.lineEditConfigRawDir,
            "rom_dir": self.ui.lineEditConfigRomDir,
            "dat_dir": self.ui.lineEditConfigDatDir,
            "parsed_dat_dir": self.ui.lineEditConfigParsedDatDir,
            "dupe_dir": self.ui.lineEditConfigDupeDir,
            "cache_dir": self.ui.lineEditConfigCacheDir,
            "log_dir": self.ui.lineEditConfigLogDir,
        }

        self.all_dirs_buttons = {
            "raw_dir": self.ui.pushButtonConfigRawDir,
            "rom_dir": self.ui.pushButtonConfigRomDir,
            "dat_dir": self.ui.pushButtonConfigDatDir,
            "parsed_dat_dir": self.ui.pushButtonConfigParsedDatDir,
            "dupe_dir": self.ui.pushButtonConfigDupeDir,
            "cache_dir": self.ui.pushButtonConfigCacheDir,
            "log_dir": self.ui.pushButtonConfigLogDir,
        }

        # Get the buttons hooked up to the line edits. Because we're looping, lambda can get wonky so use partial
        for b in self.all_dirs_buttons:
            self.all_dirs_buttons[b].clicked.connect(partial(self.set_directory_name, line_edit=self.all_dirs[b]))

        self.romdownloader_text_fields = {
            "remote_name": self.ui.lineEditConfigRomDownloaderRemoteName,
        }

        self.discord_text_fields = {
            "webhook_url": self.ui.lineEditConfigDiscordWebhookUrl,
        }

        self.all_romsearch_options = {
            "run_romdownloader": self.ui.checkBoxConfigRunRomDownloader,
            "run_datparser": self.ui.checkBoxConfigRunDatParser,
            "run_dupeparser": self.ui.checkBoxConfigRunDupeParser,
            "run_romchooser": self.ui.checkBoxConfigRunRomChooser,
            "run_rommover": self.ui.checkBoxConfigRunRomMover,
            "dry_run": self.ui.checkBoxConfigDryRun,
        }

        self.all_romdownloader_options = {
            "sync_all": self.ui.checkBoxConfigRomDownloaderSyncAll,
            "dry_run": self.ui.checkBoxConfigRomDownloaderDryRun,
        }

        self.all_dupeparser_options = {
            "use_dat": self.ui.checkBoxConfigDupeParserUseDat,
            "use_retool": self.ui.checkBoxConfigDupeParserUseRetool,
        }

        self.all_gamefinder_options = {
            "filter_dupes": self.ui.checkBoxConfigGameFinderFilterDupes,
        }

        self.all_romparser_options = {
            "use_dat": self.ui.checkBoxConfigRomParserUseDat,
            "use_filename": self.ui.checkBoxConfigRomParserUseFilename,
        }

        self.all_romchooser_options = {
            "use_best_version": self.ui.checkBoxConfigRomChooserUseBestVersion,
            "allow_multiple_regions": self.ui.checkBoxConfigRomChooserAllowMultipleRegions,
            "filter_regions": self.ui.checkBoxConfigRomChooserFilterRegions,
            "filter_languages": self.ui.checkBoxConfigRomChooserFilterLanguages,
            "dry_run": self.ui.checkBoxConfigRomChooserDryRun,
        }

        self.all_dat_filters = {
            "add-ons": self.ui.checkBoxConfigRomChooserDatFiltersAddons,
            "applications": self.ui.checkBoxConfigRomChooserDatFiltersApplications,
            "audio": self.ui.checkBoxConfigRomChooserDatFiltersAudio,
            "bad_dumps": self.ui.checkBoxConfigRomChooserDatFiltersBadDumps,
            "console": self.ui.checkBoxConfigRomChooserDatFiltersConsole,
            "bonus_discs": self.ui.checkBoxConfigRomChooserDatFiltersBonusDiscs,
            "coverdiscs": self.ui.checkBoxConfigRomChooserDatFiltersCoverdiscs,
            "demos": self.ui.checkBoxConfigRomChooserDatFiltersDemos,
            "educational": self.ui.checkBoxConfigRomChooserDatFiltersEducational,
            "games": self.ui.checkBoxConfigRomChooserDatFiltersGames,
            "manuals": self.ui.checkBoxConfigRomChooserDatFiltersManuals,
            "multimedia": self.ui.checkBoxConfigRomChooserDatFiltersMultimedia,
            "pirate": self.ui.checkBoxConfigRomChooserDatFiltersPirate,
            "preproduction": self.ui.checkBoxConfigRomChooserDatFiltersPreproduction,
            "promotional": self.ui.checkBoxConfigRomChooserDatFiltersPromotional,
            "unlicensed": self.ui.checkBoxConfigRomChooserDatFiltersUnlicensed,
            "video": self.ui.checkBoxConfigRomChooserDatFiltersVideo,
        }

        # If we have a saved config already, load it here
        self.saved_config_json_filename = "saved_config.json"
        if os.path.exists(self.saved_config_json_filename):
            self.logger.info(f"Loading in previously saved config")

            json = load_json(self.saved_config_json_filename)
            saved_filename = json["saved_config"]
            self.load_config(saved_filename)
            self.ui.lineEditConfigConfigFile.setText(saved_filename)

    def get_platforms_changed(self, item):
        """Update platforms on check

        Will update the dictionary, and add/remove items from the include/exclude table as appropriate
        """

        platform = item.text()

        checked_state = item.checkState()
        if checked_state == Qt.CheckState.Checked:
            self.add_include_exclude_tab(platform)
            self.platform_items[platform]["check_state"] = True
        else:
            self.remove_include_exclude_tab(platform)
            self.platform_items[platform]["check_state"] = False

    def remove_include_exclude_tab(self, platform):
        """Removes an include/exclude tab from the tab widget

        Args:
            platform: Platform to delete from the table
        """

        try:
            idx = self.ui.tabWidgetConfigIncludeExclude.indexOf(self.include_exclude_dict[platform]["tab"])
        except KeyError:
            return False

        # Remove the entry from the dictionary as well
        del self.include_exclude_dict[platform]

        self.ui.tabWidgetConfigIncludeExclude.removeTab(idx)

        return True

    def add_include_exclude_tab(self,
                                platform,
                                ):
        """Add an include/exclude tab for a platform

        Args:
            platform (str): Platform to add the includes/excludes for
        """

        tab_config = QWidget()
        grid_layout = QGridLayout()
        hlayout = QHBoxLayout(tab_config)

        text_edit_include, include_layout = add_include_exclude_layout(tab_config, text=self.tr("Includes"))
        grid_layout.addLayout(include_layout, 0, 0, 1, 1)

        text_edit_exclude, exclude_layout = add_include_exclude_layout(tab_config, text=self.tr("Excludes"))
        grid_layout.addLayout(exclude_layout, 0, 2, 1, 1)

        # Middle sep
        include_exclude_sep = QFrame(tab_config)
        include_exclude_sep.setFrameShadow(QFrame.Shadow.Plain)
        include_exclude_sep.setFrameShape(QFrame.Shape.VLine)

        grid_layout.addWidget(include_exclude_sep, 0, 1, 1, 1)

        hlayout.addLayout(grid_layout)
        self.ui.tabWidgetConfigIncludeExclude.addTab(tab_config, platform)

        self.include_exclude_dict[platform] = {
            "tab": tab_config,
            "includes": text_edit_include,
            "excludes": text_edit_exclude
        }

        return True

    def populate_list(self,
                      default_config_key,
                      list_widget,
                      check_state=None,
                      ):
        """Load items from a dict, potentially checked/unchecked"""

        item_dict = {}

        all_keys = self.default_config[default_config_key]

        for key in all_keys:
            item = add_item_to_list(list_widget,
                                    self.tr(key),
                                    check_state=check_state,
                                    )
            item_dict[key] = {"item": item,
                              "check_state": check_state,
                              }

        return item_dict

    def load_config_from_menu(self):
        """Load config from the menu item"""

        filename = self.get_config_name()
        self.load_config(filename)

    def get_config_name(self):

        filename, _ = QFileDialog.getOpenFileName(self,
                                                  caption=self.tr("Open config"),
                                                  dir=os.getcwd(),
                                                  filter=self.tr("Config file (*.yml;*.yaml)"),
                                                  )
        self.ui.lineEditConfigConfigFile.setText(filename)

        return filename

    def load_config(self,
                    filename,
                    ):
        """Load config yaml from file"""

        # Only load if we actually have a filename
        if filename != "":
            self.config = load_yml(filename)

            # Start by setting lists, then checkboxes, then text fields, then any extraneous things left
            self.set_lists()
            self.set_checkboxes()
            self.set_text_fields()
            self.set_romsearch_method()
            self.set_includes_excludes()

    def set_directory_name(self,
                           line_edit,
                           ):
        """Make a button set a directory name

        Args:
            line_edit (QLineEdit): The QLineEdit widget to set the text for
        """

        filename = QFileDialog.getExistingDirectory(self,
                                                    caption=self.tr("Select directory"),
                                                    dir=os.getcwd(),
                                                    )
        if filename != "":
            line_edit.setText(filename)

    def set_lists(self):
        """Set the various lists across the config"""

        self.set_platforms()
        self.set_regions()
        self.set_languages()

    def set_platforms(self):
        """Set the platform list"""

        if "platforms" not in self.config:
            return False

        platforms = self.config["platforms"]
        set_ordered_list(self.ui.listWidgetConfigPlatforms,
                         item_dict=self.platform_items,
                         items=platforms,
                         )

    def set_regions(self):
        """Set the regions list"""

        if "region_preferences" not in self.config:
            return False

        regions = self.config["region_preferences"]
        set_ordered_list(self.ui.listWidgetConfigRegionsLanguagesRegions,
                         item_dict=self.region_items,
                         items=regions,
                         )

    def set_languages(self):
        """Set the language list"""

        if "language_preferences" not in self.config:
            return False

        languages = self.config["language_preferences"]
        set_ordered_list(self.ui.listWidgetConfigRegionsLanguagesLanguages,
                         item_dict=self.language_items,
                         items=languages,
                         )

    def set_checkboxes(self):
        """Set checkboxes across the config"""

        keys = {
            "romsearch": self.all_romsearch_options,
            "romdownloader": self.all_romdownloader_options,
            "dupeparser": self.all_dupeparser_options,
            "gamefinder": self.all_gamefinder_options,
            "romparser": self.all_romparser_options,
            "romchooser": self.all_romchooser_options,
        }

        # Set all the various tabs
        for k in keys:
            self.set_generic_checkbox(k, keys[k])

        # Dat filters we do a bit differently
        self.set_dat_filter_checkboxes()

    def set_generic_checkbox(self,
                             config_key,
                             options,
                             ):
        """Set checkboxes across generic tabs in the config"""

        if config_key not in self.config:
            return False

        for o in options:
            if o in self.config[config_key]:
                check_state = self.config[config_key][o]

                options[o].setChecked(check_state)

        return True

    def set_dat_filter_checkboxes(self):
        """Set the .dat filter checkboxes"""

        if "romchooser" not in self.config:
            return False

        # We also want to loop over the filter dats here, if they're not in the list set them False
        if "dat_filters" in self.config["romchooser"]:

            all_filters = self.config["romchooser"]["dat_filters"]
            if isinstance(all_filters, str):
                all_filters = [all_filters]

            for d in self.all_dat_filters:
                if d in all_filters:
                    self.all_dat_filters[d].setChecked(True)
                else:
                    self.all_dat_filters[d].setChecked(False)

        return True

    def set_text_fields(self):
        """Set checkboxes across the config"""

        keys = {
            "dirs": self.all_dirs,
            "romdownloader": self.romdownloader_text_fields,
            "discord": self.discord_text_fields,
        }

        for k in keys:
            self.set_generic_text_fields(k, keys[k])

    def set_generic_text_fields(self,
                                config_key,
                                text_fields,
                                ):
        """Set generic text fields

        Args:
            config_key (str): The config key to set text fields for
            text_fields: Dictionary of the various text fields to set values for
        """

        if config_key not in self.config:
            return False

        for d in text_fields:
            if d in self.config[config_key]:
                field_val = self.config[config_key][d]
                text_fields[d].setText(field_val)

    def set_romsearch_method(self):
        """Set the ROMSearch method from the config file"""

        if "romsearch" not in self.config:
            return False

        if "method" in self.config["romsearch"]:

            method = self.config["romsearch"]["method"]

            if method == "filter_then_download":
                button = self.ui.radioButtonConfigRomsearchMethodFilterDownload
            elif method == "download_then_filter":
                button = self.ui.radioButtonConfigRomsearchMethodDownloadFilter
            else:
                raise ValueError(f"Do not understand ROMSearch method {method}")

            button.setChecked(True)

        return True

    def set_includes_excludes(self):
        """Set any includes/excludes config file"""

        mappings = {
            "include_games": "includes",
            "exclude_games": "excludes",
        }

        for m in mappings:
            if m in self.config:

                platforms = self.config[m]
                for p in platforms:
                    includes = "\n".join(self.config[m][p])

                    self.include_exclude_dict[p][mappings[m]].setPlainText(includes)

        return True

    def create_file(self):
        """Get filename to save config file to"""

        filename, _ = QFileDialog.getSaveFileName(self,
                                                  caption=self.tr("New config"),
                                                  dir=os.getcwd(),
                                                  filter=self.tr("Config file (*.yml)"),
                                                  )
        self.ui.lineEditConfigConfigFile.setText(filename)

    def save_config(self):
        """Get config values and save to file"""

        # Get the filename
        filename = self.ui.lineEditConfigConfigFile.text()

        # Build up the dictionary
        self.config = {}

        # Get the config directories
        self.get_generic_text_value("dirs",
                                    self.all_dirs,
                                    )

        # Platforms
        self.get_ordered_list(self.ui.listWidgetConfigPlatforms,
                              "platforms")

        # Regions/languages
        self.get_ordered_list(self.ui.listWidgetConfigRegionsLanguagesRegions,
                              "region_preferences")
        self.get_ordered_list(self.ui.listWidgetConfigRegionsLanguagesLanguages,
                              "language_preferences")

        # Includes/Excludes
        self.get_include_excludes()

        keys = {
            "romsearch": self.all_romsearch_options,
            "romdownloader": self.all_romdownloader_options,
            "dupeparser": self.all_dupeparser_options,
            "gamefinder": self.all_gamefinder_options,
            "romparser": self.all_romparser_options,
            "romchooser": self.all_romchooser_options,
        }

        # Module options. Checkboxes first
        self.get_romsearch_method()
        for k in keys:
            self.get_generic_options(k, keys[k])

        # Then text fields
        keys = {
            "romdownloader": self.romdownloader_text_fields,
            "discord": self.discord_text_fields,
        }
        for k in keys:
            self.get_generic_text_value(k, keys[k])

        # Dat filters we do a bit differently
        self.get_dat_filters()

        # Save out the config file
        save_yml(filename, self.config)

        # Also save out this to somewhere accessible, so we know it exists
        json = {"saved_config": filename}
        save_json(json, self.saved_config_json_filename)

    def get_config_dirs(self):
        """Get directories from the main tab"""

        if "dirs" not in self.config:
            self.config["dirs"] = {}

        for d in self.all_dirs:
            dir_name = self.all_dirs[d].text()
            if dir_name != "":
                self.config["dirs"][d] = dir_name

    def get_generic_text_value(self,
                               key,
                               text_fields,
                               add_at_top=True,
                               ):
        """Get text fields generically

        Args:
            key (str): The config key to get text fields for
            text_fields (dict): Dictionary of the various text fields to get values for
            add_at_top (bool): If True, add the text fields at the top of the dictionary.
                Defaults to True
        """

        if key not in self.config:
            self.config[key] = {}

        config = {}

        for d in text_fields:
            text_val = text_fields[d].text()
            if text_val != "":
                config[d] = text_val

        if add_at_top:
            self.config[key] = {**config, **self.config[key]}
        else:
            self.config[key] = {**self.config[key], **config}

        # If the dictionary is empty, delete
        if len(self.config[key]) == 0:
            del self.config[key]

    def get_ordered_list(self,
                         list_widget,
                         key,
                         ):
        """Get checked items from an ordered list"""

        n_items = list_widget.count()

        items = []

        for i in range(n_items):

            # Get item name, and check state
            item = list_widget.item(i)
            item_name = item.text()
            check_state = item.checkState()

            if check_state == Qt.CheckState.Checked:
                items.append(item_name)

        # Only add if we actually have some
        if len(items) > 0:
            self.config[key] = items

    def get_include_excludes(self):
        """Get includes/excludes from the list"""

        # Includes
        self.add_include_exclude_to_config(config_key="include_games",
                                           dict_key="includes",
                                           )
        self.add_include_exclude_to_config(config_key="exclude_games",
                                           dict_key="excludes",
                                           )

    def add_include_exclude_to_config(self,
                                      config_key,
                                      dict_key,
                                      ):
        """Add from the include/exclude text edit boxes to config

        Args:
            config_key (str): The key for the config dict
            dict_key (str): The key for the include/exclude dict
        """

        self.config[config_key] = {}

        for platform in self.include_exclude_dict:

            all_vals = self.include_exclude_dict[platform][dict_key].toPlainText()
            all_vals_list = all_vals.split("\n")

            # Loop over, find any potential blank entries
            blank_idx = []
            for i in all_vals_list:
                if i == "":
                    blank_idx.append(i)
            for idx in blank_idx[::-1]:
                all_vals_list.remove(idx)

            # Only add if we have something
            if len(all_vals_list) > 0:
                self.config[config_key][platform] = all_vals_list

        if len(self.config[config_key]) == 0:
            del self.config[config_key]

        return True

    def get_generic_options(self,
                            key,
                            options,
                            ):
        """Get generic options from a config tab

        Args:
            key (str): The config key to get fields for
            options (dict): Dictionary of the various options to get values for
        """

        if key not in self.config:
            self.config[key] = {}

        for o in options:
            checked_state = options[o].isChecked()
            self.config[key][o] = checked_state

        # If the dictionary is empty, delete
        if len(self.config[key]) == 0:
            del self.config[key]

        return True

    def get_romsearch_method(self):
        """Get the ROMSearch method from the button config"""

        if "romsearch" not in self.config:
            self.config["romsearch"] = {}

        # Get the method from the radio button group
        romsearch_method_text = self.ui.radioButtonConfigRomsearchMethod.checkedButton().text()

        if romsearch_method_text == "Filter, then download":
            romsearch_method = "filter_then_download"
        elif romsearch_method_text == "Download, then filter":
            romsearch_method = "download_then_filter"
        else:
            raise ValueError(f"Do not understand ROMSearch method {romsearch_method_text}")

        self.config["romsearch"]["method"] = romsearch_method

        return True

    def get_dat_filters(self):
        """Set the dat filters as a list"""

        if "romchooser" not in self.config:
            self.config["romchooser"] = {}

        # dat filters
        dat_filters = []
        for d in self.all_dat_filters:
            checked_state = self.all_dat_filters[d].isChecked()
            if checked_state:
                dat_filters.append(d)
        if len(dat_filters) > 0:
            self.config["romchooser"]["dat_filters"] = dat_filters

        return True



import copy
import os
import re

import numpy as np

import romsearch
from ..util import (setup_logger,
                    load_yml,
                    get_parent_name,
                    get_short_name,
                    load_json
                    )


def get_all_games(files,
                  default_config=None,
                  regex_config=None,
                  ):
    """Get all unique game names from a list of game files."""

    games = [get_short_name(f, default_config=default_config, regex_config=regex_config) for f in files]
    games = list(np.unique(games))

    return games


def get_priority(dupe_dict, parent_name, game_name):
    """Get priority from a dupe dictionary"""

    # First case: parent name doesn't exist in the dupe dict
    if parent_name not in dupe_dict:
        return 1

    # Second case: it does (potentially can be lowercase)
    dupes = [dupe.lower() for dupe in dupe_dict[parent_name]]
    reg_dupes = [dupe for dupe in dupe_dict[parent_name]]

    if game_name.lower() in dupes:
        found_parent_idx = dupes.index(game_name.lower())
        priority = dupe_dict[parent_name][reg_dupes[found_parent_idx]]["priority"]

        return priority

    # Otherwise, just return 1
    return 1


class GameFinder:

    def __init__(self,
                 platform,
                 config_file=None,
                 config=None,
                 default_config=None,
                 regex_config=None,
                 logger=None,
                 ):
        """Tool to find games within a list of files

        Will parse through files to get a unique list of games, then pull
        out potential aliases and optionally remove things from user excluded list
        
        Args:
            platform (str): Platform name
            config_file (str, optional): Path to config file. Defaults to None.
            config (dict, optional): Configuration dictionary. Defaults to None.
            default_config (dict, optional): Default configuration dictionary. Defaults to None.
            regex_config (dict, optional): Dictionary of regex config. Defaults to None.
            logger (logging.Logger, optional): Logger instance. Defaults to None.
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger = setup_logger(log_level="info",
                                  script_name=f"GameFinder",
                                  log_dir=log_dir,
                                  additional_dir=platform,
                                  )
        self.logger = logger

        # Pull in specifics to include/exclude
        self.include_games = self.config.get("include_games", None)
        self.exclude_games = self.config.get("exclude_games", None)

        # Pull in defaults for finding short game names
        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        # Info for dupes
        self.dupe_dir = config.get("dirs", {}).get("dupe_dir", None)
        self.filter_dupes = config.get("gamefinder", {}).get("filter_dupes", True)

    def run(self,
            files,
            ):

        games_dict = self.get_game_dict(files)
        games_dict = dict(sorted(games_dict.items()))

        self.logger.debug(f"Found {len(games_dict)} games:")
        for g in games_dict:
            self.logger.debug(f"{g}: {games_dict[g]}")

        return games_dict

    def get_game_dict(self,
                      files,
                      ):

        games = get_all_games(files,
                              default_config=self.default_config,
                              regex_config=self.regex_config,
                              )

        # We need to trim down dupes here. Otherwise, the
        #  dict is just the list we already have
        game_dict = None
        if self.filter_dupes:
            game_dict = self.get_filter_dupes(games)

        # If the dupe filtering has failed, then just assume everything is unique
        if game_dict is None:

            game_dict = {}

            for game in games:
                game_dict[game] = {"priority": 1,
                                   }

        # Remove any excluded files
        if self.exclude_games is not None:
            games_to_remove = self.get_game_matches(game_dict,
                                                    self.exclude_games,
                                                    )
            if games_to_remove is not None:
                for g in games_to_remove:
                    del game_dict[g]

        # Include only included files
        if self.include_games is not None:

            games_to_include = self.get_game_matches(game_dict,
                                                     self.include_games,
                                                     )
            if games_to_include is not None:

                filtered_game_dict = {}
                for g in games_to_include:
                    filtered_game_dict[g] = game_dict[g]

                game_dict = copy.deepcopy(filtered_game_dict)

        return game_dict

    def get_game_matches(self,
                         game_dict,
                         games_to_match,
                         ):
        """Get files that match an input dictionary (so as to properly handle dupes

        Args:
            - game_dict (dict): Dictionary of games to match against
            - games_to_match (list): List of values to match against
        """
        games_matched = []

        if isinstance(games_to_match, dict):
            games_to_match = games_to_match.get(self.platform, None)
        else:
            games_to_match = copy.deepcopy(games_to_match)

        if games_to_match is None:
            return None

        games_matched.extend(games_to_match)

        game_dict_keys = []
        for g in game_dict:
            found_f = False

            for game_matched in games_matched:

                if found_f:
                    continue

                # Look in the group name
                re_find = re.findall(f"^({re.escape(game_matched)}).*", g)

                if len(re_find) > 0:
                    game_dict_keys.append(g)
                    found_f = True

                # If not found, look in the dupe names
                if not found_f:
                    for g_d in game_dict[g]:

                        if found_f:
                            continue

                        re_find = re.findall(f"^({re.escape(game_matched)}).*", g_d)

                        if len(re_find) > 0:
                            game_dict_keys.append(g)
                            found_f = True

        game_dict_keys = np.unique(game_dict_keys)

        if len(game_dict_keys) == 0:
            game_dict_keys = None

        return game_dict_keys

    def get_filter_dupes(self, games):
        """Parse down a list of files based on an input dupe list"""

        if self.dupe_dir is None:
            raise ValueError("dupe_dir must be specified if filtering dupes")

        dupe_file = os.path.join(self.dupe_dir, f"{self.platform} (dupes).json")
        if not os.path.exists(dupe_file):
            self.logger.warning("No dupe files found")
            return None

        game_dict = {}

        dupes = load_json(dupe_file)

        # Loop over games, and the dupes dictionary. Also pull out priority
        for g in games:

            found_parent_name = get_parent_name(game_name=g,
                                                dupe_dict=dupes,
                                                )

            found_parent_name_lower = found_parent_name.lower()
            game_dict_keys = [key for key in game_dict.keys()]
            game_dict_keys_lower = [key.lower() for key in game_dict.keys()]

            if found_parent_name_lower not in game_dict_keys_lower:
                game_dict[found_parent_name] = {}
                final_parent_name = copy.deepcopy(found_parent_name)
            else:
                final_parent_idx = game_dict_keys_lower.index(found_parent_name_lower)
                final_parent_name = game_dict_keys[final_parent_idx]

            priority = get_priority(dupe_dict=dupes, parent_name=found_parent_name, game_name=g)

            game_dict[final_parent_name][g] = {"priority": priority}

        return game_dict



import logging
import os
import sys
from logging.handlers import RotatingFileHandler

import colorlog
from pathvalidate import sanitize_filename


def setup_logger(log_level,
                 script_name,
                 log_dir,
                 additional_dir="",
                 max_logs=9,
                 ):
    """
    Set up the logger.

    Parameters:
        log_level (str): The log level to use
        script_name (str): The name of the script
        log_dir (str): The directory to save logs to
        additional_dir (str): Any additional directories to keep log files tidy
        max_logs (int): Maximum number of log files to keep

    Returns:
        A logger object for logging messages.
    """

    # Sanitize the directories if we need to
    additional_dir = [sanitize_filename(f) for f in additional_dir.split(os.path.sep)]

    if os.environ.get('DOCKER_ENV'):
        log_dir = os.path.join(log_dir, script_name, *additional_dir)
    else:
        log_dir = os.path.join(log_dir, script_name, *additional_dir)

    if log_level not in ['debug', 'info', 'critical']:
        log_level = 'info'
        print(f"Invalid log level '{log_level}', defaulting to 'info'")

    # Create the log directory if it doesn't exist
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # Define the log file path, and sanitize if needs be
    log_file = os.path.join(log_dir, f"{script_name}.log")

    # Check if log file already exists
    if os.path.isfile(log_file):
        for i in range(max_logs - 1, 0, -1):
            old_log = f"{log_dir}/{script_name}.log.{i}"
            new_log = f"{log_dir}/{script_name}.log.{i + 1}"
            if os.path.exists(old_log):
                if os.path.exists(new_log):
                    os.remove(new_log)
                os.rename(old_log, new_log)
        os.rename(log_file, os.path.join(log_dir, f"{script_name}.log.1"))

    # Create a logger object with the script name
    logger = logging.getLogger(script_name)
    logger.propagate = False

    # Set the log level based on the provided parameter
    log_level = log_level.upper()
    if log_level == 'DEBUG':
        logger.setLevel(logging.DEBUG)
    elif log_level == 'INFO':
        logger.setLevel(logging.INFO)
    elif log_level == 'CRITICAL':
        logger.setLevel(logging.CRITICAL)
    else:
        logger.critical(f"Invalid log level '{log_level}', defaulting to 'INFO'")
        logger.setLevel(logging.INFO)

    # Define the log message format for the log files
    logfile_formatter = logging.Formatter(fmt='%(asctime)s %(levelname)s: %(message)s', datefmt='%m/%d/%y %I:%M %p')

    # Create a RotatingFileHandler for log files
    handler = RotatingFileHandler(log_file, delay=True, mode="w", backupCount=max_logs)
    handler.setFormatter(logfile_formatter)

    # Add the file handler to the logger
    logger.addHandler(handler)

    # Configure console logging with the specified log level
    console_handler = colorlog.StreamHandler(sys.stdout)
    if log_level == 'DEBUG':
        console_handler.setLevel(logging.DEBUG)
    elif log_level == 'INFO':
        console_handler.setLevel(logging.INFO)
    elif log_level == 'CRITICAL':
        console_handler.setLevel(logging.CRITICAL)

    # Add the console handler to the logger
    console_handler.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(levelname)s: %(message)s"))
    logger.addHandler(console_handler)

    # Overwrite previous logger if exists
    logging.getLogger(script_name).handlers.clear()
    logging.getLogger(script_name).addHandler(handler)
    logging.getLogger(script_name).addHandler(console_handler)

    return logger



import copy
import glob
import os
import subprocess

import romsearch
from ..util import (setup_logger,
                    load_yml,
                    get_file_pattern,
                    discord_push,
                    split,
                    )


def add_rclone_filter(pattern=None,
                      filter_type="include",
                      include_wildcard=True,
                      ):
    if filter_type == "include":
        filter_str = "+"
    elif filter_type == "exclude":
        filter_str = "-"
    else:
        raise ValueError("filter_type should be one of include or exclude")

    # rclone wants double curly braces which we need to escape in python strings (yum)
    filter_pattern = ""

    if pattern is not None:
        filter_pattern += f"{{{{{pattern}}}}}"

    if include_wildcard:
        filter_pattern += "*"

    cmd = f' --filter "{filter_str} {filter_pattern}"'

    return cmd


def get_tidy_files(glob_pattern):
    """Get a tidy list of files from a glob pattern.

    This just strips off the leading directories to just get a filename

    Args:
        glob_pattern (str): glob pattern to match
    """

    files = glob.glob(glob_pattern)
    files = [os.path.split(f)[-1] for f in files]

    return files


class ROMDownloader:

    def __init__(self,
                 platform=None,
                 config_file=None,
                 config=None,
                 platform_config=None,
                 logger=None,
                 override_includes=None,
                 override_excludes=None,
                 include_filter_wildcard=True,
                 ):
        """Downloader tool via rclone

        This works per-platform, so must be specified here

        Args:
            platform (str, optional): Platform name. Defaults to None, which will throw a ValueError
            config (str, optional): Configuration file. Defaults to None
            config (dict, optional): Configuration dictionary. Defaults to None
            platform_config (dict, optional): Platform configuration dictionary. Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None
            override_includes (list, optional): If set, will override the config includes with custom
                ones. Defaults to None.
            override_excludes (list, optional): If set, will override the config excludes with custom
                ones. Defaults to None.
            include_filter_wildcard (bool, optional): If set, will include wildcards in rclone filters. Defaults to
                True.
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger = setup_logger(log_level="info",
                                  script_name=f"ROMDownloader",
                                  log_dir=log_dir,
                                  additional_dir=platform,
                                  )
        self.logger = logger

        out_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if out_dir is None:
            raise ValueError("raw_dir needs to be defined in config")
        self.out_dir = os.path.join(out_dir, platform)

        # Get any specific includes/excludes
        include_games = self.config.get("include_games", None)
        if isinstance(include_games, dict):
            include_games = include_games.get(platform, None)
        else:
            include_games = copy.deepcopy(include_games)

        if override_includes is not None:
            include_games = copy.deepcopy(override_includes)

        self.include_games = include_games

        exclude_games = self.config.get("exclude_games", None)
        if isinstance(exclude_games, dict):
            exclude_games = exclude_games.get(platform, None)
        else:
            exclude_games = copy.deepcopy(exclude_games)

        if override_excludes is not None:
            exclude_games = copy.deepcopy(override_excludes)

        self.exclude_games = exclude_games

        remote_name = self.config.get("romdownloader", {}).get("remote_name", None)
        if remote_name is None:
            raise ValueError("remote_name must be specified in config")
        self.remote_name = remote_name

        sync_all = self.config.get("romdownloader", {}).get("sync_all", True)

        # If we have includes or excludes, force sync all False
        if self.include_games is not None or self.exclude_games is not None:
            sync_all = False

        self.sync_all = sync_all

        self.include_filter_wildcard = include_filter_wildcard

        # Read in the specific platform configuration
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(mod_dir, "configs", "platforms", f"{platform}.yml")
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        ftp_dir = self.platform_config.get("ftp_dir", None)
        if ftp_dir is None:
            raise ValueError(f"ftp_dir should be defined in the platform config file!")
        self.ftp_dir = ftp_dir

        self.discord_url = self.config.get("discord", {}).get("webhook_url", None)
        self.dry_run = self.config.get("romdownloader", {}).get("dry_run", False)

    def run(self,
            ):
        """Run Rclone sync tool"""

        start_files = get_tidy_files(os.path.join(str(self.out_dir), "*"))

        self.rclone_sync(ftp_dir=self.ftp_dir,
                         out_dir=self.out_dir,
                         )

        end_files = get_tidy_files(os.path.join(str(self.out_dir), "*"))

        if self.discord_url is not None:
            name = f"ROMDownloader: {self.platform}"
            self.post_to_discord(start_files,
                                 end_files,
                                 name=name
                                 )

        # If there are potential additional files to download, do that here
        if "additional_dirs" in self.platform_config:

            for add_dir in self.platform_config["additional_dirs"]:
                add_ftp_dir = self.platform_config["additional_dirs"][add_dir]

                add_out_dir = f"{self.out_dir} {add_dir}"

                start_files = get_tidy_files(os.path.join(str(add_out_dir), "*"))

                self.rclone_sync(ftp_dir=add_ftp_dir,
                                 out_dir=add_out_dir,
                                 )

                end_files = get_tidy_files(os.path.join(str(add_out_dir), "*"))

                if self.discord_url is not None:
                    name = f"ROMDownloader: {self.platform} ({add_dir})"
                    self.post_to_discord(start_files,
                                         end_files,
                                         name=name
                                         )

    def rclone_sync(self,
                    ftp_dir,
                    out_dir=None,
                    transfers=5,
                    ):

        if out_dir is None:
            out_dir = os.getcwd()

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        cmd = f'rclone sync -P --transfers {transfers} "{self.remote_name}:{ftp_dir}" "{out_dir}"'

        # We mostly do full syncs here, but we can specify specific game names
        if not self.sync_all:

            # Start with any negative filters
            searches = []

            if self.exclude_games is not None:
                searches.extend(self.exclude_games)

            if len(searches) > 0:
                pattern = get_file_pattern(searches)
            else:
                pattern = None

            if pattern:
                cmd += add_rclone_filter(pattern=pattern,
                                         filter_type="exclude",
                                         include_wildcard=self.include_wildcard,
                                         )

            # Now onto positive filters
            searches = []

            # Specific games
            if self.include_games is not None:
                searches.extend(self.include_games)

            if len(searches) > 0:
                pattern = get_file_pattern(searches)
            else:
                pattern = None

            if pattern:
                cmd += add_rclone_filter(pattern=pattern,
                                         filter_type="include",
                                         include_wildcard=self.include_filter_wildcard,
                                         )

                cmd += ' --filter "- *"'

        if self.dry_run:
            self.logger.info(f"Dry run, would rclone_sync with:")
            self.logger.info(cmd)
        else:

            if not os.path.exists(self.out_dir):
                os.makedirs(self.out_dir)

            # Execute the command and capture the output
            with subprocess.Popen(cmd, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as process:
                for line in process.stdout:
                    # Log each line of the output using the provided logger
                    self.logger.info(line[:-1])  # Exclude the newline character

        return True

    def post_to_discord(self,
                        start_files,
                        end_files,
                        name,
                        max_per_message=10,
                        ):
        """Create a discord post summarising files added and removed

        Args:
            start_files (list): list of files at the start of the rclone
            end_files (list): list of files at the end of the rclone
            name (string): Name of the post title
            max_per_message (int, optional): Maximum number of items per post. Defaults to 10.
        """

        items_added = list(set(end_files).difference(start_files))
        items_deleted = list(set(start_files).difference(end_files))

        if len(items_added) > 0:

            items_added.sort()

            for items_split in split(items_added, chunk_size=max_per_message):

                fields = []

                field_dict = {"name": "Added",
                              "value": "\n".join(items_split)
                              }
                fields.append(field_dict)

                if len(fields) > 0:
                    discord_push(url=self.discord_url,
                                 name=name,
                                 fields=fields,
                                 )

        if len(items_deleted) > 0:

            items_deleted.sort()

            for items_split in split(items_deleted, chunk_size=max_per_message):

                fields = []

                field_dict = {"name": "Deleted",
                              "value": "\n".join(items_split)
                              }
                fields.append(field_dict)

                if len(fields) > 0:
                    discord_push(url=self.discord_url,
                                 name=name,
                                 fields=fields,
                                 )

        return True



import copy
import glob
import os
import re

import numpy as np

import romsearch
from .datparser import DATParser
from .dupeparser import DupeParser
from .gamefinder import GameFinder
from .romchooser import ROMChooser
from .romdownloader import ROMDownloader
from .rommover import ROMMover
from .romparser import ROMParser
from ..util import (load_yml,
                    load_json,
                    setup_logger,
                    discord_push,
                    split,
                    get_short_name,
                    get_file_time,
                    )

ALLOWED_ROMSEARCH_METHODS = [
    "filter_then_download",
    "download_then_filter",
]


class ROMSearch:

    def __init__(self,
                 config_file=None,
                 config=None,
                 default_config=None,
                 regex_config=None,
                 logger=None,
                 ):
        """General search tool to get ROMs downloaded and organized into files

        Args:
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            default_config (dict, optional): default configuration dictionary. Defaults to None.
            regex_config (dict, optional): regex configuration dictionary. Defaults to None.
        """

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        # Pull in directories from the yaml file
        self.raw_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if self.raw_dir is None:
            raise ValueError("raw_dir needs to be defined in config")

        self.rom_dir = self.config.get("dirs", {}).get("rom_dir", None)
        if self.rom_dir is None:
            raise ValueError("rom_dir needs to be defined in config")

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger = setup_logger("info", script_name="ROMSearch", log_dir=log_dir)
        self.logger = logger

        # Read in the various pre-set configs we've got
        self.mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(self.mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(self.mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        # Pull out platforms, make sure they're all valid
        platforms = self.config.get("platforms", None)
        if platforms is None:
            platforms = []
        if isinstance(platforms, str):
            platforms = [platforms]
        for platform in platforms:
            if platform not in self.default_config["platforms"]:
                raise ValueError(f"Platforms should be any of {self.default_config['platforms']}, not {platform}")
        self.platforms = platforms

        self.romsearch_method = self.config.get("romsearch", {}).get("method", "filter_then_download")

        # Which modules to run
        self.run_romdownloader = self.config.get("romsearch", {}).get("run_romdownloader", True)
        self.run_datparser = self.config.get("romsearch", {}).get("run_datparser", True)
        self.run_dupeparser = self.config.get("romsearch", {}).get("run_dupeparser", True)
        self.run_romchooser = self.config.get("romsearch", {}).get("run_romchooser", True)
        self.run_rommover = self.config.get("romsearch", {}).get("run_rommover", True)
        self.dry_run = self.config.get("romsearch", {}).get("dry_run", False)

        # Finally, the discord URL if we're sending messages
        self.discord_url = self.config.get("discord", {}).get("webhook_url", None)

    def run(self):
        """Run ROMSearch"""

        self.logger.info(f"Looping over platforms: {', '.join(self.platforms)}")

        for platform in self.platforms:

            self.logger.info(f"Running ROMSearch for {platform}")

            # Pull in platform-specific config
            platform_config_file = os.path.join(self.mod_dir, "configs", "platforms", f"{platform}.yml")
            platform_config = load_yml(platform_config_file)

            raw_dir = os.path.join(self.raw_dir, platform)

            # Parse DAT files here, if we're doing that
            if self.run_datparser:
                dat_parser = DATParser(platform=platform,
                                       config=self.config,
                                       platform_config=platform_config,
                                       logger=self.logger,
                                       )
                dat_parser.run()

            # Get dupes here, if we're doing that
            if self.run_dupeparser:
                dupe_parser = DupeParser(platform=platform,
                                         config=self.config,
                                         default_config=self.default_config,
                                         regex_config=self.regex_config,
                                         logger=self.logger,
                                         )
                dupe_parser.run()

            if self.romsearch_method == "download_then_filter":
                # Run the rclone sync
                if self.run_romdownloader:
                    downloader = ROMDownloader(platform=platform,
                                               config=self.config,
                                               platform_config=platform_config,
                                               logger=self.logger,
                                               )
                    downloader.run()

                # Get the original directory, so we can safely move back after
                orig_dir = os.getcwd()
                os.chdir(raw_dir)

                all_files = glob.glob("*.zip")
                all_files.sort()

                os.chdir(orig_dir)

            elif self.romsearch_method == "filter_then_download":

                if not self.run_datparser:
                    raise ValueError("If using filter, then download method, you must run DATParser")

                parsed_dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
                if parsed_dat_dir is None:
                    raise ValueError("parsed_dat_dir needs to be defined in config")

                parsed_dat_file = os.path.join(parsed_dat_dir, f"{platform} (dat parsed).json")
                parsed_dat = load_json(parsed_dat_file)

                all_files = [f"{f}.zip" for f in parsed_dat]
                all_files = np.unique(all_files)
                all_files.sort()

            else:

                raise ValueError(f"ROMSearch method should be one of {ALLOWED_ROMSEARCH_METHODS}")

            # Parse this into a dictionary with some useful info for each file
            all_file_dict = {}
            for f in all_files:
                short_name = get_short_name(f,
                                            regex_config=self.regex_config,
                                            default_config=self.default_config,
                                            )

                all_file_dict[f] = {
                    "short_name": short_name,
                    "matched": False,
                }

            # Find files
            finder = GameFinder(platform=platform,
                                config=self.config,
                                default_config=self.default_config,
                                regex_config=self.regex_config,
                                logger=self.logger,
                                )

            all_games = finder.run(files=all_file_dict)
            self.logger.info(f"Searching through {len(all_games)} game(s)")

            all_roms_moved = []
            all_roms_dict = {}

            for i, game in enumerate(all_games):

                self.logger.info(f"{i + 1}/{len(all_games)}: {game} (aliases {', '.join(all_games[game])})")

                # Parse by the short name, include the priority in there as well
                rom_files = {}

                games_lower = [g.lower() for g in all_games[game]]
                priorities = [all_games[game][g]["priority"] for g in all_games[game]]
                for f in all_file_dict:
                    file_short_name_lower = all_file_dict[f]["short_name"].lower()
                    if file_short_name_lower in games_lower:
                        games_idx = games_lower.index(file_short_name_lower)
                        rom_files[f] = {
                            "priority": priorities[games_idx],
                        }

                        if all_file_dict[f]["matched"]:
                            raise ValueError(f"{f} has already been matched! This should not happen")

                        all_file_dict[f]["matched"] = True

                parse = ROMParser(platform=platform,
                                  game=game,
                                  config=self.config,
                                  platform_config=platform_config,
                                  regex_config=self.regex_config,
                                  default_config=self.default_config,
                                  logger=self.logger,
                                  )
                rom_dict = parse.run(rom_files)

                if self.run_romchooser:
                    # Here, we'll parse down the number of files to one game, one ROM
                    chooser = ROMChooser(platform=platform,
                                         game=game,
                                         config=self.config,
                                         regex_config=self.regex_config,
                                         default_config=self.default_config,
                                         logger=self.logger,
                                         )
                    rom_dict = chooser.run(rom_dict)

                if len(rom_dict) == 0:
                    self.logger.info(f"All files filtered. Skipping")
                    continue

                # Print out all the ROMs we've now matched
                rom_files = [f for f in rom_dict]
                self.logger.info(f"Found ROM file(s): {', '.join(rom_files)}")

                # Save to a big dictionary, since we'll move all at once
                all_roms_dict[game] = rom_dict

            for f in all_file_dict:
                if not all_file_dict[f]["matched"]:
                    self.logger.debug(f"{f} not matched to anything")

            if self.dry_run:
                self.logger.info("Dry run, will not move any files")
                continue

            if not self.run_rommover:
                self.logger.debug("ROMMover is not running, will not move anything")
                continue

            # If we filter then download, this is where we download
            if self.romsearch_method == "filter_then_download":
                all_files = []
                for game in all_roms_dict:

                    # Make sure we're using the regex escaped version for all the strings
                    fs = [re.escape(f) for f in all_roms_dict[game]]

                    all_files.extend(fs)

                if self.run_romdownloader:
                    downloader = ROMDownloader(platform=platform,
                                               config=self.config,
                                               platform_config=platform_config,
                                               logger=self.logger,
                                               override_includes=all_files,
                                               override_excludes=[],
                                               include_filter_wildcard=False,
                                               )
                    downloader.run()

                # Replace the file time with the correct one on disk
                for game in all_roms_dict:

                    for f in all_roms_dict[game]:
                        full_filename = os.path.join(self.raw_dir, platform, f)

                        file_mod_time = get_file_time(full_filename,
                                                      datetime_format=self.default_config["datetime_format"],
                                                      )
                        all_roms_dict[game][f]["file_mod_time"] = file_mod_time

            for game in all_roms_dict:

                rom_dict = all_roms_dict[game]

                mover = ROMMover(platform=platform,
                                 game=game,
                                 config=self.config,
                                 platform_config=platform_config,
                                 logger=self.logger,
                                 )
                roms_moved = mover.run(rom_dict)
                all_roms_moved.extend(roms_moved)

            # Post these to Discord in chunks of 10
            if self.discord_url is not None and len(all_roms_moved) > 0:

                for items_split in split(all_roms_moved):

                    fields = []

                    field_dict = {"name": platform,
                                  "value": "\n".join(items_split)
                                  }
                    fields.append(field_dict)

                    if len(fields) > 0:
                        discord_push(url=self.discord_url,
                                     name="ROMSearch",
                                     fields=fields,
                                     )

        self.logger.info("Complete!")

        return True



import os

import numpy as np
import requests

import romsearch
from ..util import (setup_logger,
                    load_yml,
                    get_parent_name,
                    get_short_name,
                    load_json,
                    save_json,
                    )

ID_CLONE_KEYS = [
    "cloneof",
    "cloneofid",
]


class DupeParser:

    def __init__(self,
                 platform=None,
                 config_file=None,
                 config=None,
                 default_config=None,
                 regex_config=None,
                 logger=None,
                 ):
        """Tool for figuring out a list of dupes

        Args:
            platform (str, optional): Platform name. Defaults to None, which will throw a ValueError.
            config_file (str, optional): Path to config file. Defaults to None
            config (dict, optional): Configuration dictionary. Defaults to None
            default_config (dict, optional): Default configuration dictionary. Defaults to None
            regex_config (dict, optional): Configuration dictionary for regex search. Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None

        TODO:
            - At some point, we might want to consider adding in the retool supersets
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger = setup_logger(log_level="info",
                                  script_name=f"DupeParser",
                                  log_dir=log_dir,
                                  additional_dir=platform,
                                  )
        self.logger = logger

        self.use_dat = self.config.get("dupeparser", {}).get("use_dat", True)
        self.use_retool = self.config.get("dupeparser", {}).get('use_retool', True)

        self.parsed_dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
        if self.use_dat and self.parsed_dat_dir is None:
            raise ValueError("Must specify parsed_dat_dir if using dat files")

        self.dupe_dir = self.config.get("dirs", {}).get("dupe_dir", None)
        if self.dupe_dir is None:
            raise ValueError("dupe_dir should be specified in config file")

        # Pull in platform config that we need
        mod_dir = os.path.dirname(romsearch.__file__)
        retool_config_file = os.path.join(mod_dir, "configs", "clonelists", f"retool.yml")
        retool_config = load_yml(retool_config_file)

        self.retool_url = retool_config.get("url", None)
        self.retool_platform_file = retool_config.get(platform, None)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

    def run(self):
        """Run the dupe parser"""

        if (self.retool_platform_file is None or self.retool_url is None) and self.use_retool:
            self.logger.warning("retool config for the platform needs to be present if using retool")
            return False

        dupe_dict = self.get_dupe_dict()

        # Save out the dupe dict
        out_file = os.path.join(self.dupe_dir, f"{self.platform} (dupes).json")
        save_json(dupe_dict, out_file)

        return True

    def get_dupe_dict(self):
        """Loop through potentially both the dat files and the retool config file to get out dupes"""

        dupe_dict = {}

        # Prefer retool dupes first
        if self.use_retool:
            dupe_dict = self.get_retool_dupes(dupe_dict)
        if self.use_dat:
            dupe_dict = self.get_dat_dupes(dupe_dict)

        dupe_dict = dict(sorted(dupe_dict.items()))

        return dupe_dict

    def get_dat_dupes(self, dupe_dict=None):
        """Get dupes from the dat that we've already parsed to JSON"""

        if dupe_dict is None:
            dupe_dict = {}

        json_dat = os.path.join(self.parsed_dat_dir, f"{self.platform} (dat parsed).json")
        if not os.path.exists(json_dat):
            self.logger.warning(f"No dat file found for {self.platform}")
            return None

        self.logger.info(f"Using parsed dat file {json_dat}")

        dat_dict = load_json(json_dat)

        all_keys = list(dat_dict.keys())

        for clone_name in dat_dict:
            for id_clone_key in ID_CLONE_KEYS:
                if id_clone_key in dat_dict[clone_name]:
                    clone_key = dat_dict[clone_name][id_clone_key]

                    # If it's an ID, find that ID
                    if id_clone_key == "cloneofid":

                        # Sometimes, IDs are missing from the dat so just move on
                        try:
                            dat_idx = np.where([dat_dict[key]["id"] == clone_key for key in dat_dict])[0][0]
                        except IndexError:
                            continue
                        parent_entry = dat_dict[all_keys[dat_idx]]
                        parent_name = parent_entry["name"]

                    elif id_clone_key == "cloneof":
                        # TODO
                        raise NotImplemented("Only current implemented for cloneofid")
                    else:
                        raise ValueError(f"Only know how to parse {ID_CLONE_KEYS}")

                    # Get short names here
                    # parent_game_name = get_game_name(parent_name)
                    parent_game_name = get_short_name(parent_name,
                                                      default_config=self.default_config,
                                                      regex_config=self.regex_config,
                                                      )
                    clone_short_name = get_short_name(clone_name,
                                                      default_config=self.default_config,
                                                      regex_config=self.regex_config,
                                                      )

                    # If the names are the same, just skip
                    if parent_game_name == clone_short_name:
                        continue

                    found_parent_name = get_parent_name(game_name=parent_game_name,
                                                        dupe_dict=dupe_dict,
                                                        )
                    if found_parent_name not in dupe_dict:
                        dupe_dict[found_parent_name] = {}

                    dupe_dict[found_parent_name][clone_short_name] = {"priority": 1}

        return dupe_dict

    def get_retool_dupes(self, dupe_dict=None):
        """Get dupes from the retool curated list"""

        if dupe_dict is None:
            dupe_dict = {}

        retool_dupes = self.get_retool_dupe_dict()
        for retool_dupe in retool_dupes:

            # If we don't have titles within the dupe dict, skip
            if "titles" not in retool_dupe:
                continue

            group = retool_dupe["group"]
            group_titles = [get_short_name(f["searchTerm"],
                                           default_config=self.default_config,
                                           regex_config=self.regex_config,
                                           )
                            for f in retool_dupe["titles"]]
            priorities = [f.get("priority", 1) for f in retool_dupe["titles"]]

            group_parsed = get_short_name(group,
                                          default_config=self.default_config,
                                          regex_config=self.regex_config,
                                          )

            found_parent_name = get_parent_name(game_name=group_parsed,
                                                dupe_dict=dupe_dict,
                                                )
            if found_parent_name not in dupe_dict:
                dupe_dict[found_parent_name] = {}

            for i, g in enumerate(group_titles):
                dupe_dict[found_parent_name][g] = {"priority": priorities[i]}

        return dupe_dict

    def download_retool_dupe(self,
                             out_file=None,
                             just_date=False,
                             ):
        """Download the retool curated list, optionally just returning the last modified date"""

        retool_url = f"{self.retool_url}/{self.retool_platform_file}"
        with requests.get(retool_url) as r:
            retool_dict = r.json()
            if just_date:
                return retool_dict["description"]["lastUpdated"]
            retool_full_file = r.text

        if out_file is None:
            raise ValueError("Should specify an out_file to save the retool dupe list to")

        with open(out_file, "w", encoding="utf-8") as f:
            f.write(retool_full_file)

        return True

    def get_retool_dupe_dict(self):
        """Pull the retool duplicates out of the clonelist file"""

        if not os.path.exists(self.dupe_dir):
            os.makedirs(self.dupe_dir)

        retool_dupe_file = os.path.join(self.parsed_dat_dir, f"{self.platform} (retool).json")
        if not os.path.exists(retool_dupe_file):

            if not os.path.exists(self.parsed_dat_dir):
                os.makedirs(self.parsed_dat_dir)

            self.logger.info("No retool dupe file found. Downloading")
            self.download_retool_dupe(retool_dupe_file)

        retool_dupes = load_json(retool_dupe_file)

        # Check if there's a more updated file, if so download it
        local_file_time = retool_dupes["description"]["lastUpdated"]
        remote_file_time = self.download_retool_dupe(just_date=True)

        if not local_file_time == remote_file_time:
            self.logger.info("More up-to-date dupe file found. Will download")
            self.download_retool_dupe(retool_dupe_file)

        self.logger.info(f"Using retool clonelist {retool_dupe_file}")

        retool_dupes = load_json(retool_dupe_file)
        retool_dupes = retool_dupes["variants"]

        return retool_dupes



import sys

from PySide6.QtWidgets import QApplication

from romsearch.gui import MainWindow


def main():
    app = QApplication(sys.argv)

    window = MainWindow()
    window.show()

    app.exec()


if __name__ == '__main__':
    main()



import copy
import os
from collections import Counter

import numpy as np
import packaging.version
from packaging import version

import romsearch
from ..util import (setup_logger,
                    load_yml,
                    )

DAT_FILTERS = [
    "add-ons",
    "applications",
    "audio",
    "bad_dumps",
    "console",
    "bonus_discs",
    "coverdiscs",
    "demos",
    "educational",
    "games",
    "manuals",
    "multimedia",
    "pirate",
    "preproduction",
    "promotional",
    "unlicensed",
    "video",
]


def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)


def remove_rom_dict_entries(rom_dict,
                            key,
                            remove_type="bool",
                            bool_remove=True,
                            list_preferences=None,
                            ):
    """Remove entries from the game dict based on various rules"""

    f_to_delete = []
    for f in rom_dict:

        # Bool type, we can filter either on Trues or Falses
        if remove_type == "bool":

            key_val = rom_dict[f].get(key, None)
            if key_val is None:
                continue

            if bool_remove:
                if key_val:
                    f_to_delete.append(f)
            else:
                if not key_val:
                    f_to_delete.append(f)

        elif remove_type == "list":

            if list_preferences is None:
                raise ValueError("list_preferences not specified")

            # If there's no information in there, assume we're OK
            if len(rom_dict[f][key]) == 0:
                continue

            found = False
            for val in rom_dict[f][key]:
                if val in list_preferences:
                    found = True
            if not found:
                f_to_delete.append(f)

        else:
            raise ValueError("remove_type should be one of bool, list")

    f_to_delete = np.unique(f_to_delete)
    for f in f_to_delete:
        rom_dict.pop(f)

    return rom_dict


def get_best_version(rom_dict,
                     version_key="version",
                     ):
    """Pull out all the regions we've got left, to loop over and search for versions"""

    all_regions = np.unique([",".join(rom_dict[key]["regions"]) for key in rom_dict])

    for region in all_regions:
        region_rom_dict = {key: rom_dict[key][version_key]
                           for key in rom_dict if ",".join(rom_dict[key]["regions"]) == region}

        # Pull out all the versions we have
        all_vers = [region_rom_dict[key] for key in region_rom_dict]
        # If we have anything here that doesn't have a version, set it to v0
        all_vers = [vers if vers != "" else "v0" for vers in all_vers]

        # If we have lettered versions, convert these to numbers
        for i, vers in enumerate(all_vers):
            try:
                version.parse(vers)
            except packaging.version.InvalidVersion:
                all_vers[i] = f"v{ord(vers[1:])}"

        all_keys = [key for key in region_rom_dict]

        max_ver = max(all_vers, key=version.parse)
        max_ver_idx = np.where(np.asarray(all_vers) == max_ver)[0]

        max_ver_key = np.asarray(all_keys)[max_ver_idx]

        keys_to_pop = []
        for key in all_keys:
            if key not in max_ver_key:
                keys_to_pop.append(key)

        for key in keys_to_pop:
            rom_dict.pop(key)

    return rom_dict


def get_best_rom_per_region(rom_dict,
                            region_preferences,
                            ):
    """For each individual region, get an overall best ROM"""
    for reg_pref in region_preferences:

        roms = []

        for key in rom_dict:
            if reg_pref in rom_dict[key]["regions"]:
                roms.append(key)

        if len(roms) > 1:
            roms = get_best_roms(roms, rom_dict)

            keys_to_pop = []
            for f in rom_dict:
                if f not in roms and reg_pref in rom_dict[f]["regions"]:
                    keys_to_pop.append(f)

            for key in keys_to_pop:
                rom_dict.pop(key)

    return rom_dict


def remove_unwanted_roms(rom_dict, key_to_check, check_type="include"):
    """Remove unwanted ROMs from the dict

    If we have multiple versions lying around that may be preferred or demoted for some reason, parse them
    out here. Do this per region combo
    """

    all_regions = []
    for key in rom_dict:
        all_regions.extend(",".join(rom_dict[key]["regions"]))
    all_regions = np.unique(all_regions)

    keys_to_pop = []

    for region in all_regions:

        region_game_keys = [key for key in rom_dict if region in rom_dict[key]["regions"]]

        # Only filter things down if we've got multiples here
        if len(region_game_keys) > 1:
            found_demoted = [rom_dict[key][key_to_check] for key in region_game_keys]

            # Remove these, but only if we have some but not all
            if 0 < sum(found_demoted) < len(found_demoted):
                for key in region_game_keys:
                    if check_type == "include":
                        if not rom_dict[key][key_to_check]:
                            keys_to_pop.append(key)
                    elif check_type == "exclude":
                        if rom_dict[key][key_to_check]:
                            keys_to_pop.append(key)
                    else:
                        raise ValueError(f"check_type should be one of include or exclude")

    for key in keys_to_pop:
        rom_dict.pop(key)

    return rom_dict


def get_best_roms(files, rom_dict):
    """Get the best ROM(s) from a list, using a scoring system"""

    # Positive scores
    improved_version_score = 1
    version_score = 10
    revision_score = 100
    budget_edition_score = 1000

    # Negative scores
    demoted_version_score = -1
    alternate_version_score = -1
    modern_version_score = -10
    priority_score = -100

    file_scores = np.zeros(len(files))

    # Improved or demoted versions
    file_scores += improved_version_score * np.array([int(rom_dict[f]["improved_version"]) for f in files])
    file_scores += demoted_version_score * np.array([int(rom_dict[f]["demoted_version"]) for f in files])
    file_scores += budget_edition_score * np.array([int(rom_dict[f]["budget_edition"]) for f in files])
    file_scores += alternate_version_score * np.array([int(rom_dict[f]["alternate"]) for f in files])

    # Version numbering, which needs to be parsed
    file_scores += version_score * add_versioned_score(files, rom_dict, "version")
    file_scores += revision_score * add_versioned_score(files, rom_dict, "revision")

    # Downweight modern versions
    file_scores += modern_version_score * np.array([int(rom_dict[f]["modern_version"]) for f in files])

    # Priority scoring, where we subtract 1 so that the highest priority has no changed
    file_scores += priority_score * (np.array([int(rom_dict[f]["priority"]) for f in files]) - 1)

    files_idx = np.where(file_scores == np.nanmax(file_scores))[0]
    files = np.asarray(files)[files_idx]

    return files


def add_versioned_score(files, rom_dict, key):
    """Get an order for versioned strings"""

    # Ensure we have a version here. If blank, set to v0
    rom_dict = copy.deepcopy(rom_dict)
    for f in rom_dict:
        if rom_dict[f][key] == "":
            rom_dict[f][key] = "v0"

    versions = np.array([version.parse(rom_dict[f][key]) for f in files])
    versions_clean = [key for key, value in Counter(versions).most_common()]
    version_vals = sorted(range(len(versions_clean)), key=versions.__getitem__)

    versions_sorted = versions[version_vals]

    file_scores_version = np.zeros(len(files))
    for i, v in enumerate(versions_sorted):
        v_idx = np.where(versions == v)[0]
        file_scores_version[v_idx] += i

    return file_scores_version


def filter_by_list(rom_dict,
                   key,
                   key_prefs,
                   ):
    """Find file with highest value in given list. If there are multiple matches, find the most updated one"""

    roms = []

    found_key = False
    for key_pref in key_prefs:

        if found_key:
            continue

        for val in rom_dict:
            if key_pref in rom_dict[val][key]:
                roms.append(val)
                found_key = True

    # If we have multiple matches here, define a best match using scoring system
    if len(roms) > 1:
        roms = get_best_roms(roms,
                             rom_dict,
                             )

    keys_to_pop = []
    for f in rom_dict:
        if f not in roms:
            keys_to_pop.append(f)

    for key in keys_to_pop:
        rom_dict.pop(key)

    return rom_dict


class ROMChooser:

    def __init__(self,
                 platform,
                 game,
                 config_file=None,
                 config=None,
                 default_config=None,
                 regex_config=None,
                 logger=None,
                 ):
        """ROM choose tool

        This works per-game, per-platform, so must be specified here

        Args:
            platform (str): Platform name
            game (str): Game name
            config_file (str, optional): Path to config file. Defaults to None.
            config (dict, optional): Configuration dictionary. Defaults to None.
            default_config (dict, optional): Default configuration dictionary. Defaults to None.
            regex_config (dict, optional): Configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): Logger instance. Defaults to None.
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger_add_dir = str(os.path.join(platform, game))
            logger = setup_logger(log_level="info",
                                  script_name=f"ROMChooser",
                                  log_dir=log_dir,
                                  additional_dir=logger_add_dir,
                                  )
        self.logger = logger

        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        # Region preference (usually set USA for retroachievements, can also be a list to fall back to)
        region_preferences = self.config.get("region_preferences", self.default_config["default_region"])
        if isinstance(region_preferences, str):
            region_preferences = [region_preferences]

        for region_pref in region_preferences:
            if region_pref not in self.default_config["regions"]:
                raise ValueError(f"Regions should be any of {self.default_config['regions']}, not {region_pref}")

        self.region_preferences = region_preferences

        # Language preference (usually set En, can also be a list to fall back to)
        language_preferences = self.config.get("language_preferences", self.default_config["default_language"])
        if isinstance(language_preferences, str):
            language_preferences = [language_preferences]

        for language_pref in language_preferences:
            if language_pref not in self.default_config["languages"]:
                raise ValueError(f"Regions should be any of {self.default_config['languages']}, not {language_pref}")

        self.language_preferences = language_preferences

        # Various filters. First are the boolean ones
        dat_filters = self.config.get("romchooser", {}).get("dat_filters", "all_but_games")
        if "all" in dat_filters:
            all_dat_filters = copy.deepcopy(DAT_FILTERS)
            if "all_but" in dat_filters:
                filter_to_remove = dat_filters.split("all_but_")[-1]
                all_dat_filters.remove(filter_to_remove)
            dat_filters = copy.deepcopy(all_dat_filters)

        if isinstance(dat_filters, str):
            dat_filters = [dat_filters]
        self.dat_filters = dat_filters

        self.filter_regions = self.config.get("romchooser", {}).get("filter_regions", True)
        self.filter_languages = self.config.get("romchooser", {}).get("filter_languages", True)
        self.allow_multiple_regions = self.config.get("romchooser", {}).get("allow_multiple_regions", False)
        self.use_best_version = self.config.get("romchooser", {}).get("use_best_version", True)

        self.dry_run = self.config.get("romchooser", {}).get("dry_run", False)

    def run(self,
            rom_dict):
        """Run the ROM chooser"""

        rom_dict = self.run_chooser(rom_dict)

        return rom_dict

    def run_chooser(self,
                    rom_dict):
        """Make a ROM choice based on various factors

        This chooser works in this order:

        - Removing any demo files
        - Removing any beta files
        - Removing anything where the language isn't in the user preferences
            (for files with no language info, this will skipped)
        - Removing anything where the region isn't in the user preferences
        - Get some "best version", via:
            - Revision number
            - Version number
            - Some kind of special name to indicate an improved version
        - Finally, if we only allow one region, parse down to a single region (first in the list)
        """

        for f in self.dat_filters:
            self.logger.debug(f"Filtering {f}")
            if f in DAT_FILTERS:
                rom_dict = remove_rom_dict_entries(rom_dict,
                                                   f,
                                                   remove_type="bool",
                                                   bool_remove=True,
                                                   )
            else:
                raise ValueError(f"Unknown filter type {f}")

        # Language
        if self.filter_languages:
            self.logger.debug("Filtering languages")
            rom_dict = remove_rom_dict_entries(rom_dict,
                                               "languages",
                                               remove_type="list",
                                               list_preferences=self.language_preferences,
                                               )

        # Regions
        if self.filter_regions:
            self.logger.debug("Filtering regions")
            rom_dict = remove_rom_dict_entries(rom_dict,
                                               "regions",
                                               remove_type="list",
                                               list_preferences=self.region_preferences,
                                               )

        # Remove versions we potentially don't want around
        if self.use_best_version:
            self.logger.debug("Getting best version")
            rom_dict = get_best_version(rom_dict)
            rom_dict = remove_unwanted_roms(rom_dict, key_to_check="improved_versions", check_type="include")
            rom_dict = remove_unwanted_roms(rom_dict, key_to_check="budget_edition", check_type="include")
            rom_dict = remove_unwanted_roms(rom_dict, key_to_check="demoted_versions", check_type="exclude")
            rom_dict = remove_unwanted_roms(rom_dict, key_to_check="modern_versions", check_type="exclude")
            rom_dict = remove_unwanted_roms(rom_dict, key_to_check="alternate", check_type="exclude")
            rom_dict = get_best_rom_per_region(rom_dict,
                                               self.region_preferences,
                                               )

        if not self.allow_multiple_regions:
            self.logger.debug("Trimming down to a single region")
            rom_dict = filter_by_list(rom_dict,
                                      "regions",
                                      self.region_preferences,
                                      )

        return rom_dict



from importlib.metadata import version

# Get the version
__version__ = version(__name__)

from .modules import (DATParser,
                      DupeParser,
                      GameFinder,
                      ROMDownloader,
                      ROMChooser,
                      ROMMover,
                      ROMParser,
                      ROMSearch
                      )

__all__ = [
    "DATParser",
    "DupeParser",
    "GameFinder",
    "ROMDownloader",
    "ROMChooser",
    "ROMMover",
    "ROMParser",
    "ROMSearch",
]



import copy
import os
import re

import romsearch
from ..util import (setup_logger,
                    get_file_time,
                    load_yml,
                    load_json,
                    get_game_name,
                    )

DICT_DEFAULT_VALS = {
    "bool": False,
    "str": "",
    "list": []
}


def find_pattern(regex, search_str, group_number=0):
    """
    Take a regex pattern and find potential matches within a search string
    """
    regex_search_str = None

    regex_search = re.search(regex, search_str)
    if regex_search:
        regex_search_str = regex_search.group(group_number)

    return regex_search_str


def get_pattern_val(regex,
                    tag,
                    regex_type,
                    pattern_mappings=None
                    ):
    """Get values out from a regex pattern, optionally mapping back to something more readable for lists"""

    pattern_string = find_pattern(regex, tag)

    if pattern_string is not None:
        pattern_string = pattern_string.strip("()")

        if regex_type == "bool":
            pattern_val = True
        elif regex_type == "str":
            pattern_val = pattern_string
        elif regex_type == "list":

            if pattern_mappings is not None:
                parsed_pattern_string = []
                # Match to pattern mappings
                for p in pattern_mappings:
                    if re.match(pattern_mappings[p], pattern_string):
                        parsed_pattern_string.append(p)
            else:
                # Split, and remove and trailing whitespace
                parsed_pattern_string = pattern_string.split(",")
                parsed_pattern_string = [s.strip() for s in parsed_pattern_string]
            pattern_val = parsed_pattern_string
        else:
            raise ValueError("regex_type should be one of 'bool', 'str', or 'list'")

    else:
        pattern_val = None

    return pattern_val


class ROMParser:

    def __init__(self,
                 platform,
                 game,
                 config_file=None,
                 config=None,
                 platform_config=None,
                 default_config=None,
                 regex_config=None,
                 logger=None,
                 ):
        """ROM parser tool

        This works per-game, per-platform, so must be specified here

        Args:
            platform (str): Platform name
            game (str): Game name
            config_file (str, optional): path to config file. Defaults to None.
            config (dict, optional): configuration dictionary. Defaults to None.
            platform_config (dict, optional): platform configuration dictionary. Defaults to None.
            default_config (dict, optional): default configuration dictionary. Defaults to None.
            regex_config (dict, optional): regex configuration dictionary. Defaults to None.
            logger (logging.Logger, optional): logger instance. Defaults to None.

        TODO:
            - Default implied languages from regions
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger_add_dir = str(os.path.join(platform, game))
            logger = setup_logger(log_level="info",
                                  script_name=f"ROMParser",
                                  log_dir=log_dir,
                                  additional_dir=logger_add_dir,
                                  )
        self.logger = logger

        mod_dir = os.path.dirname(romsearch.__file__)

        if default_config is None:
            default_file = os.path.join(mod_dir, "configs", "defaults.yml")
            default_config = load_yml(default_file)
        self.default_config = default_config

        if regex_config is None:
            regex_file = os.path.join(mod_dir, "configs", "regex.yml")
            regex_config = load_yml(regex_file)
        self.regex_config = regex_config

        if platform_config is None:
            platform_config_file = os.path.join(mod_dir, "configs", "platforms", f"{platform}.yml")
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.raw_dir = self.config.get("dirs", {}).get("raw_dir", None)
        if not self.raw_dir:
            raise ValueError("raw_dir must be specified in config.yml")

        self.use_dat = self.config.get("romparser", {}).get("use_dat", True)
        self.use_retool = self.config.get("romparser", {}).get("use_retool", True)
        self.use_filename = self.config.get("romparser", {}).get("use_filename", True)
        self.dry_run = self.config.get("romparser", {}).get("dry_run", False)

        # If we're using the dat file, pull it out here
        self.dat = None
        if self.use_dat:
            dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
            if dat_dir is None:
                raise ValueError("parsed_dat_dir must be specified in config.yml")
            dat_file = os.path.join(dat_dir, f"{platform} (dat parsed).json")
            if os.path.exists(dat_file):
                self.dat = load_json(dat_file)

        # If we're using the retool file, pull it out here
        self.retool = None
        if self.use_retool:
            dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)
            if dat_dir is None:
                raise ValueError("parsed_dat_dir must be specified in config.yml")
            retool_file = os.path.join(dat_dir, f"{platform} (retool).json")
            if os.path.exists(retool_file):
                self.retool = load_json(retool_file)

    def run(self,
            files,
            ):
        """Run the ROM parser"""

        game_dict = {}

        for f in files:
            game_dict[f] = self.parse_file(f)

            # Include the priority
            game_dict[f]["priority"] = files[f]["priority"]

        return game_dict

    def parse_file(self,
                   f,
                   ):
        """Parse useful info out of a specific file"""

        file_dict = {}

        if self.use_filename:
            file_dict = self.parse_filename(f, file_dict)
        if self.use_retool:
            file_dict = self.parse_retool(f, file_dict)
        if self.use_dat:
            file_dict = self.parse_dat(f, file_dict)

        # Any last minute finalisations
        self.finalise_file_dict(file_dict)

        # File modification time
        full_file_path = os.path.join(self.raw_dir, self.platform, f)
        file_time = get_file_time(full_file_path,
                                  datetime_format=self.default_config["datetime_format"],
                                  )
        file_dict["file_mod_time"] = file_time

        self.logger.debug(f"{f}: {file_dict}")

        return file_dict

    def parse_retool(self, f, file_dict=None):
        """Parse info out of the retool file"""

        if file_dict is None:
            file_dict = {}

        if self.retool is None:
            self.logger.warning(f"No retool file found for {self.platform}. Skipping")
            return file_dict

        # Pull out the game name
        game_name = get_game_name(f)

        # Loop over the variants, see if we get a match
        found_cat = False
        for retool_dict in self.retool["variants"]:

            if found_cat:
                continue

            # If we don't have titles within the dupe dict, skip
            if "titles" not in retool_dict:
                continue

            retool_variants = [f["searchTerm"].lower() for f in retool_dict["titles"]]

            if game_name.lower() in retool_variants:

                found_cat = True

                # If we have categories, set these to True
                retool_cats = retool_dict.get("categories", [])
                for retool_cat in retool_cats:
                    file_cat = retool_cat.lower().replace(" ", "_")
                    file_dict[file_cat] = True

        return file_dict

    def parse_dat(self, f, file_dict=None):
        """Parse info out of the dat file"""

        if file_dict is None:
            file_dict = {}

        if self.dat is None:
            self.logger.warning(f"No dat file found for {self.platform}. Skipping")
            return file_dict

        # Remember there aren't zips in the dat entries
        dat_entry = self.dat.get(f.strip(".zip"), None)
        if not dat_entry:
            self.logger.warning(f"No dat entry found for {f}. Skipping")
            return file_dict

        dat_categories = self.default_config.get("dat_categories", [])
        for dat_cat in dat_categories:

            dat_val = dat_entry.get("category", "")
            cat_val = dat_val == dat_cat

            dat_cat_dict = dat_cat.lower().replace(" ", "_")
            if dat_cat_dict in file_dict:
                file_dict[dat_cat_dict] = file_dict[dat_cat_dict] | cat_val
            else:
                file_dict[dat_cat_dict] = cat_val

        return file_dict

    def finalise_file_dict(self,
                           file_dict,
                           ):

        dat_categories = self.default_config.get("dat_categories", [])

        for d in dat_categories:
            d_sanitized = d.lower().replace(" ", "_")

            if d_sanitized not in file_dict:
                file_dict[d_sanitized] = False

        if all([file_dict[d.lower().replace(" ", "_")] is False for d in dat_categories]):
            file_dict["games"] = True

        return file_dict

    def parse_filename(self, f, file_dict=None):
        """Parse info out of filename"""

        if file_dict is None:
            file_dict = {}

        # Split file into tags
        tags = [f'({x}' for x in f.strip(".zip").split(' (')][1:]

        for regex_key in self.regex_config:

            regex_type = self.regex_config[regex_key].get("type", "bool")
            search_tags = self.regex_config[regex_key].get("search_tags", True)
            group = self.regex_config[regex_key].get("group", None)
            regex_flags = self.regex_config[regex_key].get("flags", "I")
            transform_pattern = self.regex_config[regex_key].get("transform_pattern", None)
            transform_repl = self.regex_config[regex_key].get("transform_repl", None)

            dict_default_val = DICT_DEFAULT_VALS.get(regex_type, None)
            if dict_default_val is None:
                raise ValueError(f"regex_type should be one of {list(DICT_DEFAULT_VALS.keys())}")

            if regex_key not in file_dict:
                file_dict[regex_key] = dict_default_val

            if regex_flags == "NOFLAG":
                regex_flags = re.NOFLAG
            elif regex_flags == "I":
                regex_flags = re.I
            else:
                raise ValueError("regex_flags should be one of 'NOFLAG', 'I'")

            pattern = self.regex_config[regex_key]["pattern"]

            pattern_mappings = None

            if regex_type == "list":

                if isinstance(self.default_config[regex_key], dict):
                    str_to_join = [self.default_config[regex_key][key] for key in self.default_config[regex_key]]
                    pattern_mappings = self.default_config[regex_key]
                else:
                    str_to_join = copy.deepcopy(self.default_config[regex_key])

                pattern = pattern.replace(f"[{regex_key}]", "|".join(str_to_join))

            regex = re.compile(pattern, flags=regex_flags)
            if search_tags:

                found_tag = False

                for tag in tags:

                    if found_tag:
                        continue

                    pattern_string = get_pattern_val(regex,
                                                     tag,
                                                     regex_type,
                                                     pattern_mappings=pattern_mappings,
                                                     )
                    if pattern_string is not None:

                        if transform_pattern is not None:

                            pattern_string = re.sub(transform_pattern, transform_repl, pattern_string)

                        file_dict[regex_key] = pattern_string
                        found_tag = True
            else:
                pattern_string = get_pattern_val(regex,
                                                 f,
                                                 regex_type,
                                                 pattern_mappings=pattern_mappings
                                                 )
                if pattern_string is not None:
                    file_dict[regex_key] = pattern_string

            # Update groups, if needed
            if group is not None:
                if group not in file_dict:
                    file_dict[group] = dict_default_val

                if regex_type == "bool":
                    file_dict[group] = file_dict[group] | file_dict[regex_key]
                elif regex_type == "str":
                    if file_dict[group] and file_dict[regex_key]:
                        raise ValueError("Can't combine multiple groups with type str")
                    else:
                        file_dict[group] += file_dict[regex_key]
                elif regex_type == "list":
                    file_dict[group].extend(file_dict[regex_key])
                else:
                    raise ValueError(f"regex_type should be one of {list(DICT_DEFAULT_VALS.keys())}")

        return file_dict



from PySide6.QtCore import Slot, Signal, QObject, QThread
from PySide6.QtWidgets import (QApplication,
                               QMainWindow,
                               )

from romsearch import ROMSearch
from .gui_about import AboutWindow
from .gui_config import ConfigWindow
from .gui_utils import open_url, get_gui_logger
from .layout_romsearch import Ui_RomSearch


@Slot()
def close_all():
    """Close the application"""

    QApplication.closeAllWindows()


class MainWindow(QMainWindow):

    def __init__(self):
        """ROMSearch Main Window"""

        super().__init__()

        self.logger = get_gui_logger()
        self.logger.warning("Do not close this window!")

        self.ui = Ui_RomSearch()
        self.ui.setupUi(self)

        # Set up the worker threads for later
        self.romsearch_thread = None
        self.romsearch_worker = None

        # Set up the config file here
        self.config_window = ConfigWindow(self.ui, logger=self.logger)

        # File menu buttons
        new_config = self.ui.actionNewConfigFile
        new_config.triggered.connect(self.config_window.create_file)

        load_config = self.ui.actionLoadConfigFile
        load_config.triggered.connect(self.config_window.load_config_from_menu)

        save_config = self.ui.actionSaveConfigFile
        save_config.triggered.connect(self.config_window.save_config)

        exit_button = self.ui.actionExit
        exit_button.triggered.connect(close_all)

        # Help menu buttons
        documentation = self.ui.actionDocumentation
        documentation.triggered.connect(lambda: open_url("https://romsearch.readthedocs.io"))

        issues = self.ui.actionIssues
        issues.triggered.connect(lambda: open_url("https://github.com/bbtufty/romsearch/issues"))

        about = self.ui.actionAbout
        about.triggered.connect(lambda: AboutWindow(self).exec())

        # Main window buttons
        run_romsearch = self.ui.pushButtonRunRomsearch
        run_romsearch.clicked.connect(self.run_romsearch)

        exit_button = self.ui.pushButtonExit
        exit_button.clicked.connect(close_all)

    @Slot()
    def run_romsearch(self):
        """Run ROMSearch given the config file"""

        # Save out the config so we know we're good
        self.config_window.save_config()

        config_file = self.ui.lineEditConfigConfigFile.text()

        # Set up a thread so the UI doesn't hang
        self.romsearch_thread = QThread()
        self.romsearch_worker = RomSearchWorker(config_file=config_file)

        self.romsearch_worker.moveToThread(self.romsearch_thread)
        self.romsearch_thread.started.connect(self.romsearch_worker.run)

        # Delete the thread once we're done
        self.romsearch_worker.finished.connect(self.romsearch_thread.quit)
        self.romsearch_worker.finished.connect(self.romsearch_worker.deleteLater)
        self.romsearch_thread.finished.connect(self.romsearch_thread.deleteLater)

        # When finished, re-enable the UI
        self.romsearch_thread.finished.connect(
            lambda: self.ui.centralwidget.setEnabled(True)
        )
        # Start the thread
        self.romsearch_thread.start()

        # Disable the UI
        self.ui.centralwidget.setEnabled(False)


class RomSearchWorker(QObject):
    finished = Signal()

    def __init__(self, config_file):
        super().__init__()

        self.config_file = config_file

    def run(self):

        rs = ROMSearch(self.config_file)
        rs.run()

        self.finished.emit()



import glob
import os
from urllib.request import urlopen

import xmltodict

import romsearch
from ..util import load_yml, setup_logger, unzip_file, save_json

ALLOWED_GROUPS = [
    "No-Intro",
    "Redump",
]


def get_dat(dat_file_name,
            ):
    """Parse the dat file to a raw dictionary from a zip file"""

    with open(dat_file_name, "r") as f:
        dat = xmltodict.parse(f.read(), attr_prefix='')

    return dat


def format_dat(dat):
    """Format dat into a nice dictionary"""

    rom_dict = {}
    rom_data = dat["datafile"]["game"]

    for rom in rom_data:
        rom_dict[rom["name"]] = rom

    return rom_dict


class DATParser:

    def __init__(self,
                 platform=None,
                 config_file=None,
                 config=None,
                 platform_config=None,
                 logger=None,
                 ):
        """Parser for dat files from Redump or No-Intro

        For Redump dats, we can download directly from the site.
        Users will have to provide their own files for No-Intro,
        since there's no good way to scrape them automatically

        Args:
            platform (str, optional): Platform name. Defaults to None, which will throw a ValueError
            config_file (str, optional): Configuration file. Defaults to None
            config (dict, optional): Configuration dictionary. Defaults to None
            platform_config (dict, optional): Platform configuration dictionary. Defaults to None
            logger (logging.Logger, optional): Logger instance. Defaults to None
        """

        if platform is None:
            raise ValueError("platform must be specified")
        self.platform = platform

        if config_file is None and config is None:
            raise ValueError("config_file or config must be specified")

        if config is None:
            config = load_yml(config_file)
        self.config = config

        if logger is None:
            log_dir = self.config.get("dirs", {}).get("log_dir", os.path.join(os.getcwd(), "logs"))
            logger = setup_logger(log_level="info",
                                  script_name=f"DATParser",
                                  log_dir=log_dir,
                                  additional_dir=platform,
                                  )
        self.logger = logger

        self.dat_dir = self.config.get("dirs", {}).get("dat_dir", None)
        self.parsed_dat_dir = self.config.get("dirs", {}).get("parsed_dat_dir", None)

        self.platform = platform

        # Read in the specific platform configuration
        mod_dir = os.path.dirname(romsearch.__file__)

        if platform_config is None:
            platform_config_file = os.path.join(mod_dir, "configs", "platforms", f"{platform}.yml")
            platform_config = load_yml(platform_config_file)
        self.platform_config = platform_config

        self.group = self.platform_config.get("group", None)
        if self.group is None:
            raise ValueError("No group name specified in platform config file")
        if self.group not in ALLOWED_GROUPS:
            raise ValueError(f"Group needs to be one of {ALLOWED_GROUPS}")

        # Pull out the platform specifics for the dats
        dat_config_file = os.path.join(mod_dir, "configs", "dats", f"{self.group.lower()}.yml")

        dat_config = load_yml(dat_config_file)
        self.dat_url = dat_config.get("url", None)
        dat_config = dat_config.get(self.platform, None)
        self.dat_config = dat_config

        # Set up the name for the file
        self.out_file = os.path.join(self.parsed_dat_dir, f"{self.platform} (dat parsed).json")

    def run(self):

        run_datparser = True

        if self.dat_dir is None:
            self.logger.warning("No dat_dir defined in config file")
            run_datparser = False
        if self.parsed_dat_dir is None:
            self.logger.warning("No parsed_dat_dir defined in config file")
            run_datparser = False
        if self.dat_config is None:
            self.logger.warning("No platform-specific dat config in the dat configuration file")
            run_datparser = False

        if run_datparser:
            self.run_datparser()

        return True

    def run_datparser(self):
        """The main meat of running the dat parser"""

        zip_file = self.get_zip_file()
        if zip_file is None:
            return False

        # Unzip the file if it doesn't already exist
        dat_file_name = zip_file.replace(".zip", ".dat")
        if not os.path.exists(dat_file_name):
            unzip_file(zip_file, self.dat_dir)

        dat = get_dat(dat_file_name)

        if dat is None:
            return False

        self.logger.info(f"Using dat file {os.path.split(dat_file_name)[-1]}")

        rom_dict = format_dat(dat)

        self.save_rom_dict(rom_dict)

    def get_zip_file(self):
        """Get zip file from the dat directory

        If this is a Redump file, we can download the latest directly from the
        site. Otherwise, you will need to download them manually
        """

        file_mapping = self.dat_config.get("file_mapping", None)
        if file_mapping is None:
            raise ValueError("No file mapping defined in dat config file")

        if self.group == "Redump":
            self.download_latest_redump_dat()

        zip_files = glob.glob(os.path.join(self.dat_dir, f"{file_mapping}*.zip"))
        zip_files.sort()

        if len(zip_files) > 1:
            self.logger.info(f"Found {len(zip_files)} zip files. Will remove all but the latest (and associated dats)")
            for z in zip_files[:-1]:
                os.remove(z)
                d = z.replace(".zip", ".dat")
                if os.path.exists(d):
                    os.remove(d)

            zip_files = glob.glob(os.path.join(self.dat_dir, f"{file_mapping}*.zip"))
            zip_files.sort()

        if len(zip_files) == 0:
            self.logger.warning(f"No zip files found. "
                                f"You need to manually download {self.group} dat files for {self.platform}")
            return None

        return zip_files[-1]

    def download_latest_redump_dat(self):
        """Download Redump zip file for the platform"""

        web_mapping = self.dat_config.get("web_mapping", None)
        if web_mapping is None:
            raise ValueError("No web mapping defined in dat config file")

        response = urlopen(f"{self.dat_url}/{web_mapping}")
        f = response.headers.get_filename()

        out_file = os.path.join(self.dat_dir, f)
        if os.path.exists(out_file):
            self.logger.info(f"{f} already downloaded, will skip")
            return True

        self.logger.info(f"Downloading {f}")
        if not os.path.exists(self.dat_dir):
            os.makedirs(self.dat_dir)

        with open(out_file, mode="wb") as d:
            d.write(response.read())

        return True

    def save_rom_dict(self,
                      rom_dict,
                      ):
        """Save the dat file parsed as a dictionary to JSON"""

        if not os.path.exists(self.parsed_dat_dir):
            os.makedirs(self.parsed_dat_dir)

        out_file = os.path.join(self.out_file)
        save_json(rom_dict, out_file)



import logging
import sys

import colorlog
from PySide6.QtCore import Slot, Qt
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QListWidgetItem


@Slot()
def open_url(url):
    """Opens a URL"""

    QDesktopServices.openUrl(url)


def add_item_to_list(item_list, item_name, check_state=None):
    """Add item to list widget, optionally setting a check state

    Args:
        item_list (QListWidgetItem): Item to add
        item_name (str): Item name
        check_state (None, True, False): If None, will not set check
            state. Otherwise, will set checked (True) or not checked
            (False). Defaults to None
    """

    item = QListWidgetItem(item_list)
    item.setText(item_name)
    if check_state is not None:
        if not check_state:
            item.setCheckState(Qt.CheckState.Unchecked)
        else:
            item.setCheckState(Qt.CheckState.Checked)

    return item


def get_gui_logger(log_level="info"):
    logger = logging.getLogger()

    # Set the log level based on the provided parameter
    log_level = log_level.upper()
    if log_level == 'DEBUG':
        logger.setLevel(logging.DEBUG)
    elif log_level == 'INFO':
        logger.setLevel(logging.INFO)
    elif log_level == 'CRITICAL':
        logger.setLevel(logging.CRITICAL)
    else:
        logger.critical(f"Invalid log level '{log_level}', defaulting to 'INFO'")
        logger.setLevel(logging.INFO)
    log_handler = colorlog.StreamHandler(sys.stdout)
    log_handler.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(levelname)s: %(message)s"))
    logger.addHandler(log_handler)

    return logger



import copy
import os
import time
from datetime import datetime


def split(full_list, chunk_size=10):
    """Split a list in chunks of size chunk_size

    Args:
        full_list (list): list to split
        chunk_size (int, optional): size of each chunk. Defaults to 10
    """

    for i in range(0, len(full_list), chunk_size):
        yield full_list[i:i + chunk_size]


def get_parent_name(game_name,
                    dupe_dict,
                    ):
    """Get the parent name recursively searching through a dupe dict"""

    # We do this by lowercase checking
    reg_dupes = [g for g in dupe_dict]
    all_dupes = [g.lower() for g in dupe_dict]

    # Pull out all the clones so we can check that way as well
    all_clones = [list(dupe_dict[key].keys()) for key in dupe_dict]

    found_dupe = False

    found_parent_name = None

    # First, just check the dupes
    if game_name.lower() in all_dupes:
        found_idx = all_dupes.index(game_name.lower())
        found_parent_name = reg_dupes[found_idx]

        found_dupe = True

    # Check all the clones within the dupes
    else:
        for i, clone in enumerate(all_clones):
            if found_dupe:
                continue

            clone = [c.lower() for c in clone]
            if game_name.lower() in clone:
                found_parent_name = reg_dupes[i]
                found_dupe = True

    if not found_dupe:
        found_parent_name = copy.deepcopy(game_name)

    if found_parent_name is None:
        raise ValueError("Could not find a parent name!")

    return found_parent_name


def get_file_time(f,
                  datetime_format,
                  ):
    """Get created file time from the file itself"""

    if os.path.exists(f):
        ti_m = os.path.getmtime(f)
        date_ti_m = datetime.strptime(time.ctime(ti_m), "%a %b %d %H:%M:%S %Y")
    else:
        date_ti_m = datetime(year=1900, month=1, day=1, hour=0, minute=0, second=0)
    date_ti_m_str = date_ti_m.strftime(format=datetime_format)

    return date_ti_m_str



import copy
import re
import romsearch
import os

from ..util import load_yml


def get_file_pattern(str_to_match):
    if isinstance(str_to_match, list):
        pattern = "|".join([f"{f}" for f in str_to_match])
    else:
        pattern = f"{str_to_match}.*"

    pattern = f"({pattern})"

    return pattern


def get_bracketed_file_pattern(str_to_match,
                               ):
    """Get file pattern to match bracketed things

    This is a little tricky since sometimes things can be matched a little unusually.
    What we do here is allow for matches of letters, commas or spaces before and after,
    but ensure there's no text immediately after the match. So "De" will match for a language
    but not "Demo"
    """

    short_pattern = "[\\s\\-a-zA-Z,0-9]*?,?\\s?{},?\\s?(?![a-zA-Z])[\\s\\-a-zA-Z,0-9]*?"

    if isinstance(str_to_match, list):
        # pattern = "|".join([f".*?{f}.*?" for f in str_to_match])
        pattern = "|".join([short_pattern.replace("{}", f) for f in str_to_match])
    else:
        # pattern = f".*?{str_to_match}.*?"
        pattern = short_pattern.replace("{}", str_to_match)

    pattern = f"\\(({pattern})\\)"

    return pattern


def get_game_name(f):
    """Get game name, which is just everything up to the first bracket"""

    f = re.findall("^.*?(?=\\s\\()", f)[0]

    return f


def get_short_name(f,
                   regex_config=None,
                   default_config=None,
                   ):
    """Get short game name from the ROM file naming convention"""

    if ".zip" in f:
        f = f.strip(".zip")

    mod_dir = os.path.dirname(romsearch.__file__)

    if default_config is None:
        default_file = os.path.join(mod_dir, "configs", "defaults.yml")
        default_config = load_yml(default_file)

    if regex_config is None:
        regex_file = os.path.join(mod_dir, "configs", "regex.yml")
        regex_config = load_yml(regex_file)

    for regex_key in regex_config:

        # If we have patterns that we do want to keep in the long name, then skip
        include_in_short_title = regex_config[regex_key].get("include_in_short_name", False)
        if include_in_short_title:
            continue

        regex_type = regex_config[regex_key].get("type", "bool")
        regex_flags = regex_config[regex_key].get("flags", "I")

        pattern = regex_config[regex_key]["pattern"]

        if regex_type == "list":

            if isinstance(default_config[regex_key], dict):
                str_to_join = [default_config[regex_key][key] for key in default_config[regex_key]]
            else:
                str_to_join = copy.deepcopy(default_config[regex_key])

            pattern = pattern.replace(f"[{regex_key}]", "|".join(str_to_join))

        if regex_flags == "NOFLAG":
            regex_flags = re.NOFLAG
        elif regex_flags == "I":
            regex_flags = re.I
        else:
            raise ValueError("regex_flags should be one of 'NOFLAG', 'I'")

        pattern = f"\\s?{pattern}"

        f = re.sub(pattern=pattern, repl="", string=f, flags=regex_flags)

    return f



from discordwebhook import Discord

def discord_push(url,
                 name,
                 fields,
                 ):
    """Post a message to Discord"""

    discord = Discord(url=url)
    discord.post(
        embeds=[
            {
                "author": {
                    "name": name,
                    "url": "https://github.com/bbtufty/romsearch",
                },
                "fields": fields
            }
        ],
    )

    return True
